import tempfile
import io
import os
import tarfile
from typing import IO

import requests

def is_within_directory(directory, target):

    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory

def safe_download_extractall(url:str, path=".", members=None, *, numeric_owner=False):
    with tempfile.TemporaryFile() as f:
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=4096): 
                f.write(chunk)
        f.seek(0)
        with tarfile.open(mode='r:gz', fileobj=f) as tar:
            for member in tar.getmembers():
                member_path = os.path.join(path, member.name)
                if not is_within_directory(path, member_path):
                    raise Exception("Attempted Path Traversal in Tar File")

            tar.extractall(path, members, numeric_owner=numeric_owner) 


# SPDX-License-Identifier: Apache-2.0
# Copyright © 2022 Wrangle Ltd

from typing import Iterator, List
import tempfile
import gzip
import shutil
import typing
import csv
import io
from requests_toolbelt.multipart.encoder import MultipartEncoder

from wrgl import diffreader
from wrgl.commit import Commit, CommitResult, Table, CommitTree
from wrgl.diff import DiffResult
from wrgl.serialize import json_loads
from wrgl.uma import UMAClient


class Repository(object):
    """Represents the HTTP API that wraps a hosted Wrgl repository"""

    def __init__(
        self, repo_uri: str, client_id: str, client_secret: str = None
    ) -> None:
        """
        :param str repo_uri: the URI of the repository
        :param str client_id: Keycloak client id
        :param str client_secret: Keycloak client secret
        """
        self._client = UMAClient(repo_uri, client_id, client_secret)

    @property
    def rpt(self) -> str:
        return self._client.rpt

    def get_refs(self) -> dict:
        """Get references as a mapping of reference name and commit checksum

        :rtype: dict
        """
        r = self._client.get("/refs/")
        obj = r.json()
        return obj["refs"]

    def get_branch(self, branch: str) -> Commit:
        """Get the head commit of a branch

        :param str branch: the name of the branch

        :rtype: Commit
        """
        r = self._client.get("/refs/heads/%s/" % branch)
        return json_loads(r.content, Commit)

    def commit(
        self,
        branch: str,
        message: str,
        file: typing.BinaryIO,
        primary_key: typing.List[str],
    ) -> CommitResult:
        """Creates a new commit

        :param str branch: name of the branch to commit under
        :param str message: commit message
        :param typing.BinaryIO file: the CSV file to commit
        :param list[str] primary_key: list of column names that make up the primary key

        :rtype: CommitResult
        """
        with tempfile.TemporaryFile() as fp:
            with gzip.open(fp, "w") as gzf:
                shutil.copyfileobj(file, gzf)

            def create_request_args():
                fp.seek(0)
                m = MultipartEncoder(
                    fields={
                        "branch": branch,
                        "message": message,
                        "primaryKey": ",".join(primary_key),
                        "file": ("data.csv.gz", fp, "text/csv"),
                    }
                )
                return {
                    "data": m,
                    "headers": {"Content-Type": m.content_type},
                }

            r = self._client.post(
                "/commits/",
                create_request_args=create_request_args,
            )
        return json_loads(r.content, CommitResult)

    def get_commit_tree(self, head: str, max_depth: int) -> CommitTree:
        """Gets commit tree

        :param str head: name of the root commit, could either be reference name or commit checksum.
        :param int max_depth: maximum depth of commit tree to fetch

        :rtype: CommitTree
        """
        r = self._client.get("/commits/", params={"head": head, "maxDepth": max_depth})
        return json_loads(r.content, CommitTree)

    def get_commit(self, commit_sum: str) -> Commit:
        """Get commit with the given checksum

        :param str commit_sum: commit checksum

        :rtype: Commit
        """
        r = self._client.get("/commits/%s/" % commit_sum)
        return json_loads(r.content, Commit)

    def get_table(self, table_sum: str) -> Table:
        """Get table with the given checksum

        :param str table_sum: table checksum

        :rtype: Table
        """
        r = self._client.get("/tables/%s/" % table_sum)
        return json_loads(r.content, Table)

    def get_blocks(
        self,
        commit: str,
        start: int = None,
        end: int = None,
        with_column_names: bool = True,
    ) -> Iterator[List[str]]:
        """Fetchs blocks as concatenated rows. Each row as a list of strings.

        Calling this with default `start`, `end`, and `with_column_names` will return the entire table.

        :param str commit: either commit checksum or reference e.g. "heads/main"
        :param int start: index of the first block to fetch. Defaults to 0.
        :param int end: index of the last block to fetch. If not set, fetch til the end.
        :param bool with_column_names: prepend column names to the resulting CSV, which in effect producing a CSV with header.

        :rtype: typing.Iterator[list[str]]
        """
        r = self._client.get(
            "/blocks/",
            params={
                "head": commit,
                "start": start,
                "end": end,
                "columns": "true" if with_column_names else "false",
            },
        )
        for row in csv.reader(io.StringIO(r.text), dialect="unix"):
            yield row

    def get_table_blocks(
        self,
        table_sum: str,
        start: int = None,
        end: int = None,
        with_column_names: bool = True,
    ) -> Iterator[List[str]]:
        """Fetchs blocks with table checksum.

        Calling this with default `start`, `end`, and `with_column_names` will return the entire table.

        :param str table_sum: table checksum
        :param int start: index of the first block to fetch. Defaults to 0.
        :param int end: index of the last block to fetch. If not set, fetch til the end.
        :param bool with_column_names: prepend column names to the resulting CSV, which in effect producing a CSV with header.

        :rtype: typing.Iterator[list[str]]
        """
        r = self._client.get(
            "/tables/%s/blocks/" % table_sum,
            params={
                "start": start,
                "end": end,
                "columns": "true" if with_column_names else "false",
            },
        )
        for row in csv.reader(io.StringIO(r.text), dialect="unix"):
            yield row

    def get_rows(self, commit: str, offsets: List[int]) -> Iterator[List[str]]:
        """Get rows at certain offsets. Each row will be returned as a list of strings.

        This is usually used in tandem with row offsets from :class:`DiffResult` to fetch changed rows.

        :param str commit: either commit checksum or reference e.g. "heads/main"
        :param list[int] offsets: the offsets of the rows to fetch

        :rtype: typing.Iterator[list[str]]
        """
        r = self._client.get(
            "/rows/",
            params={"head": commit, "offsets": ",".join([str(v) for v in offsets])},
        )
        for row in csv.reader(io.StringIO(r.text), dialect="unix"):
            yield row

    def get_table_rows(self, table_sum: str, offsets: List[int]) -> Iterator[List[str]]:
        """Get rows at certain offsets with table checksum.

        This is usually used in tandem with row offsets from :class:`DiffResult` to fetch changed rows.

        :param str table_sum: table checksum
        :param list[int] offsets: the offsets of the rows to fetch

        :rtype: typing.Iterator[list[str]]
        """
        r = self._client.get(
            "/tables/%s/rows/" % table_sum,
            params={"offsets": ",".join([str(v) for v in offsets])},
        )
        for row in csv.reader(io.StringIO(r.text), dialect="unix"):
            yield row

    def diff(self, sum1: str, sum2: str) -> DiffResult:
        """Compares two commits and returns their differences.

        :param str sum1: checksum of the first commit
        :param str sum2: checksum of the second commit

        :rtype: DiffResult
        """
        r = self._client.get("/diff/%s/%s/" % (sum1, sum2))
        return json_loads(r.content, DiffResult)

    def diff_reader(
        self, sum1: str, sum2: str, fetch_size: int = 100
    ) -> diffreader.DiffReader:
        """Compares two commits and interpret their differences.

        This method is higher level than :func:`Repository.diff` and should
        be preferred for 99% of use cases.

        :param str sum1: checksum of the first commit
        :param str sum2: checksum of the second commit

        :rtype: DiffReader
        """
        return diffreader.DiffReader(self, sum1, sum2, fetch_size)



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2022 Wrangle Ltd

import json
import re
import typing

import attr


def to_camel_case(s):
    components = s.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


word_break_pattern = re.compile(r"([a-z])([A-Z])")


def to_snake_case(s):
    return word_break_pattern.sub(r"\1_\2", s).lower()


def none_field_transformer(cls, fields):
    return [
        field if field.default != attr.NOTHING else field.evolve(default=None)
        for field in fields
    ]


def field_transformer(namespace):
    def transform(cls, fields):
        results = []
        attr.resolve_types(
            cls, globalns=namespace, localns={cls.__name__: cls}, attribs=fields
        )
        for field in fields:
            original_type = getattr(field.type, "__origin__", None)
            if type(field.type) is str and field.type == cls.__name__:
                field = field.evolve(type=cls)
            if type(field.type) is type:
                validator = attr.validators.instance_of(field.type)
            elif original_type is list or original_type is typing.List:
                validator = attr.validators.deep_iterable(
                    member_validator=attr.validators.instance_of(field.type.__args__[0])
                )
            elif original_type is dict or original_type is typing.Dict:
                validator = attr.validators.deep_mapping(
                    key_validator=attr.validators.instance_of(field.type.__args__[0]),
                    value_validator=attr.validators.instance_of(field.type.__args__[1]),
                )
            else:
                raise TypeError(
                    "unanticipated type %s (%s)"
                    % (field.type, getattr(field.type, "__origin__"))
                )
            validator = attr.validators.optional(validator)
            validator = (
                validator
                if field.validator is None
                else attr.validators.and_(field.validator, validator)
            )
            results.append(field.evolve(default=None, validator=validator))
        return results

    return transform


class _JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if attr.has(obj.__class__):
            fields = attr.fields(obj.__class__)
            return {
                to_camel_case(field.name): getattr(obj, field.name)
                for field in fields
                if getattr(obj, field.name) is not None
            }
        return super().default(obj)


def json_dumps(
    inst,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    indent=None,
    separators=None,
    default=None,
    sort_keys=False,
    **kwargs
):
    return json.dumps(
        inst,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        cls=_JSONEncoder,
        indent=indent,
        separators=separators,
        default=default,
        sort_keys=sort_keys,
        **kwargs
    )


def json_loads(s, serializer_cls):
    data = json.loads(s)
    return _deserialize(data, serializer_cls)


def _deserialize(data, serializer_cls):
    kwargs = dict()
    fields_dict = attr.fields_dict(serializer_cls)
    data = [(to_snake_case(k), v) for k, v in data.items()]
    data_stack = [
        (kwargs, k, fields_dict[k], v)
        for k, v in data
        if k in fields_dict and k != "meta"
    ]
    while len(data_stack) > 0:
        parent, name, field, value = data_stack.pop()
        if value is None:
            parent[name] = None
            continue
        if type(field.type) is type:
            if attr.has(field.type):
                parent[name] = _deserialize(value, field.type)
            else:
                parent[name] = value
        elif field.type.__origin__ is list or field.type.__origin__ is typing.List:
            el_cls = field.type.__args__[0]
            if type(value) is not list:
                raise TypeError(
                    'keyword argument "%s" should be a list of %s' % (name, el_cls)
                )
            if attr.has(el_cls):
                parent[name] = [_deserialize(e, el_cls) for e in value]
            else:
                parent[name] = value
        elif field.type.__origin__ is dict or field.type.__origin__ is typing.Dict:
            el_cls = field.type.__args__[1]
            if type(value) is not dict:
                raise TypeError(
                    'keyword argument "%s" should be a dict of %s' % (name, el_cls)
                )
            if attr.has(el_cls):
                parent[name] = {k: _deserialize(e, el_cls) for k, e in value.items()}
            else:
                parent[name] = value
        else:
            raise TypeError("unanticipated field type %s" % field.type)
    return serializer_cls(**kwargs)



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2022 Wrangle Ltd

from datetime import datetime
import typing
import platform
import _csv
from unittest import TestCase
from contextlib import contextmanager
import io
import tempfile
import pathlib
import socket
import re
import time
import os
import subprocess
import csv

from wrgl.diffreader import ColumnChanges
from wrgl.repository import Repository
from wrgl import tar
from wrgl.vcr_test import use_vcr


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return str(port)


class RepositoryTestCase(TestCase):
    maxDiff = None
    repo_uri = "http://localhost:8081"
    client_id = "wrgl-python-sdk"
    client_secret = "my-secret"

    def setUp(self):
        super().setUp()
        self.repo = Repository(self.repo_uri, self.client_id, self.client_secret)

    @contextmanager
    def commit(
        self, branch: str, message: str, primary_key: typing.List[str]
    ) -> typing.Iterator["_csv._writer"]:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            writer = csv.writer(f)
            yield writer
        try:
            with open(f.name, "rb") as f2:
                self.repo.commit(branch, message, f2, primary_key)
        finally:
            os.remove(f.name)

    @use_vcr
    def test_commit(self):
        with self.commit("test_commit", "initial commit", ["a"]) as writer:
            writer.writerow(["a", "b", "c"])
            writer.writerow(["1", "q", "w"])
            writer.writerow(["2", "a", "s"])

        refs = self.repo.get_refs()
        self.assertEqual(len(refs), 1)
        commit1 = self.repo.get_commit(refs["heads/test_commit"])
        self.assertIsInstance(commit1.time, datetime)
        commit2 = self.repo.get_branch("test_commit")
        self.assertEqual(commit1, commit2)

        b = io.StringIO()
        writer = csv.writer(b)
        writer.writerow(["a", "b", "c"])
        writer.writerow(["1", "e", "w"])
        writer.writerow(["2", "c", "s"])
        cr = self.repo.commit(
            branch="test_commit",
            message="second commit",
            file=io.BytesIO(b.getvalue().encode("utf8")),
            primary_key=["a"],
        )
        self.assertIsNotNone(cr.sum)
        self.assertIsNotNone(cr.table)

        com_tree = self.repo.get_commit_tree(cr.sum, 2)
        self.assertEqual(com_tree.sum, cr.sum)
        self.assertIn(commit1.sum, com_tree.root.parents)
        self.assertIn(commit1.sum, com_tree.root.parent_commits)
        self.assertIsNotNone(com_tree.root.parent_commits[commit1.sum].table)

        tbl = self.repo.get_table(cr.table)
        self.assertEqual(tbl.columns, ["a", "b", "c"])
        self.assertEqual(
            list(self.repo.get_blocks(cr.sum)),
            [["a", "b", "c"], ["1", "e", "w"], ["2", "c", "s"]],
        )
        self.assertEqual(
            list(self.repo.get_table_blocks(cr.table)),
            [["a", "b", "c"], ["1", "e", "w"], ["2", "c", "s"]],
        )
        self.assertEqual(list(self.repo.get_rows(cr.sum, [0])), [["1", "e", "w"]])
        self.assertEqual(
            list(self.repo.get_table_rows(cr.table, [0])), [["1", "e", "w"]]
        )
        dr = self.repo.diff(cr.sum, commit1.sum)
        self.assertTrue(dr.primary_key == dr.old_primary_key)
        self.assertGreater(len(dr.row_diff), 0)

    @use_vcr
    def test_diff_reader(self):
        with self.commit("test_diff_reader", "initial commit", ["a"]) as writer:
            writer.writerow(["a", "b", "c", "d"])
            writer.writerow(["1", "q", "w", "e"])
            writer.writerow(["2", "a", "s", "d"])
            writer.writerow(["3", "z", "x", "c"])

        with self.commit("test_diff_reader", "second commit", ["a"]) as writer:
            writer.writerow(["a", "b", "c", "e"])
            writer.writerow(["1", "q", "u", "r"])
            writer.writerow(["2", "a", "s", "f"])
            writer.writerow(["4", "y", "u", "i"])

        com_tree = self.repo.get_commit_tree("heads/test_diff_reader", 2)
        dr = self.repo.diff_reader(com_tree.sum, com_tree.root.parents[0])
        self.assertEqual(
            dr.column_changes,
            ColumnChanges(
                new_values=["a", "b", "c", "e"],
                old_values=["a", "b", "c", "d"],
                unchanged=set(["a", "b", "c"]),
                added=set(["e"]),
                removed=set(["d"]),
            ),
        )
        self.assertEqual(
            dr.pk_changes,
            ColumnChanges(
                new_values=["a"],
                old_values=["a"],
                unchanged=set(["a"]),
                added=set([]),
                removed=set([]),
            ),
        )
        self.assertEqual(len(dr.added_rows), 1)
        self.assertEqual(dr.added_rows.columns, ["a", "b", "c", "e"])
        self.assertEqual(dr.added_rows.primary_key, ["a"])
        self.assertEqual(list(dr.added_rows), [["4", "y", "u", "i"]])
        self.assertEqual(len(dr.removed_rows), 1)
        self.assertEqual(dr.removed_rows.columns, ["a", "b", "c", "d"])
        self.assertEqual(dr.removed_rows.primary_key, ["a"])
        self.assertEqual(list(dr.removed_rows), [["3", "z", "x", "c"]])
        self.assertEqual(len(dr.modified_rows), 2)
        self.assertEqual(dr.modified_rows.columns, ["a", "b", "c", "d", "e"])
        self.assertEqual(dr.modified_rows.primary_key, ["a"])
        self.assertEqual(
            list(dr.modified_rows),
            [
                [("1", "1"), ("q", "q"), ("u", "w"), (None, "e"), ("r", None)],
                [("2", "2"), ("a", "a"), ("s", "s"), (None, "d"), ("f", None)],
            ],
        )
        self.assertIsNotNone(dr.data_profile)
        self.assertEqual(dr.data_profile.old_rows_count, 3)
        self.assertEqual(dr.data_profile.new_rows_count, 3)
        self.assertEqual(len(dr.data_profile.columns), 5)

    @use_vcr
    def test_diff_reader_no_changes(self):
        with self.commit(
            "test_diff_reader_no_changes", "initial commit", ["a"]
        ) as writer:
            writer.writerow(["a", "b", "c", "d"])
            writer.writerow(["1", "q", "w", "e"])
            writer.writerow(["2", "a", "s", "d"])
            writer.writerow(["3", "z", "x", "c"])

        with self.commit(
            "test_diff_reader_no_changes", "second commit", ["a"]
        ) as writer:
            writer.writerow(["a", "b", "c", "d"])
            writer.writerow(["1", "q", "w", "e"])
            writer.writerow(["2", "a", "s", "d"])
            writer.writerow(["3", "z", "x", "c"])

        com_tree = self.repo.get_commit_tree("heads/test_diff_reader_no_changes", 2)
        dr = self.repo.diff_reader(com_tree.sum, com_tree.root.parents[0])
        self.assertEqual(
            dr.column_changes,
            ColumnChanges(
                new_values=["a", "b", "c", "d"],
                old_values=["a", "b", "c", "d"],
                unchanged=set(["a", "b", "c", "d"]),
                added=set([]),
                removed=set([]),
            ),
        )
        self.assertEqual(
            dr.pk_changes,
            ColumnChanges(
                new_values=["a"],
                old_values=["a"],
                unchanged=set(["a"]),
                added=set([]),
                removed=set([]),
            ),
        )
        self.assertIsNone(dr.added_rows)
        self.assertIsNone(dr.removed_rows)
        self.assertIsNone(dr.modified_rows)



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2022 Wrangle Ltd

from wrgl.config import Config, User, Remote, Branch, Receive, Auth, Pack
from wrgl.commit import Commit, CommitResult, CommitTree, Table
from wrgl.diff import DiffResult, RowDiff
from wrgl.repository import Repository

__all__ = [
    "Config",
    "User",
    "Remote",
    "Branch",
    "Receive",
    "Auth",
    "Pack",
    "Commit",
    "CommitResult",
    "CommitTree",
    "Table",
    "DiffResult",
    "RowDiff",
    "Repository",
]



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2022 Wrangle Ltd

import typing
import attr

from wrgl.serialize import field_transformer


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class User(object):
    """User configuration, this corresponds to the `user` section in config.yaml.
    Learn more about `configuration options`_.

    In the context of a hosted repository, this holds information about the default user that perform
    automated actions such as updating reflog whenever the repository receive a push.

    :ivar str name: default user's name
    :ivar str email: default user's email
    """

    name: str
    email: str


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Receive(object):
    """Receive configuration, this corresponds to the `receive` section in config.yaml.
    Learn more about `configuration options`_.

    :ivar bool deny_non_fast_forwards: don't allow non-fast-forward pushes, learn more `here <https://www.wrgl.co/doc/wrgl-reference/push#refspec>`_
    :ivar bool deny_deletes: don't allow ref deletion
    """

    deny_non_fast_forwards: bool
    deny_deletes: bool


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Branch(object):
    """Branch's upstream configuration, this corresponds to the `branch` section in config.yaml.
    Learn more about `configuration options`_.

    This configuration is rarely useful for a hosted repository because most hosted repositories do not have upstreams.

    :ivar str remote: upstream remote of the branch
    :ivar str merge: upstream ref of the branch
    """

    remote: str
    merge: str


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Auth(object):
    """Authentication configuration, this corresponds to the `auth` section in config.yaml.
    Learn more about `configuration options`_.

    :ivar str token_duration: how long before a JWT token given by the `/authenticate/` endpoint of Wrgld expire.
        This is a string in the format "72h3m0.5s". Tokens last for 90 days by default.
    """

    token_duration: str


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Pack(object):
    """Packfile configuration, this corresponds to the `pack` section in config.yaml.
    Learn more about `configuration options`_.

    :ivar int max_file_size: maximum packfile size in bytes
    """

    max_file_size: int


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Remote(object):
    """Remote configuration, this corresponds to the `remote` section in config.yaml.
    Learn more about `configuration options`_.

    This configuration is rarely useful in the context of hosted repositories, unless you want your upstreams to have upstreams themselves.

    :ivar str url: url of the remote
    :ivar list[str] fetch: list of refspec to automatically fetch from this upstream
    :ivar list[str] push: list of refspec to automatically push to this upstream
    :ivar bool mirror: whether this upstream should mirror the repository
    """

    url: str
    fetch: typing.List[str]
    push: typing.List[str]
    mirror: bool


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Config(object):
    """Configurations as recorded in config.yaml, which can also be read/update via the `/config/ endpoint <https://www.wrgl.co/doc/http-api#config>`_.
    Learn more about `configuration options`_.

    :ivar User user: user configuration
    :ivar dict[str, Remote] remote: mapping of remote configurations
    :ivar Receive receive: receive configurations
    :ivar dict[str, Branch] branch: branch upstreams configurations
    :ivar Auth auth: authentication configurations
    :ivar Pack pack: packfile configurations
    """

    user: User
    remote: typing.Dict[str, Remote]
    receive: Receive
    branch: typing.Dict[str, Branch]
    auth: Auth
    pack: Pack



from typing import Callable
import functools

import vcr


def use_vcr(f: Callable) -> Callable:
    @functools.wraps(f)
    def run_test(*args, **kwargs):
        with vcr.use_cassette(
            f"fixtures/vcr_cassettes/{f.__module__}/{f.__qualname__.replace('.', '/')}.yaml",
            record_mode="once",
        ):
            f(*args, **kwargs)

    return run_test



import datetime


# Helpers for parsing the result of isoformat()
def _parse_isoformat_date(dtstr):
    # It is assumed that this function will only be called with a
    # string of length exactly 10, and (though this is not used) ASCII-only
    year = int(dtstr[0:4])
    if dtstr[4] != '-':
        raise ValueError('Invalid date separator: %s' % dtstr[4])

    month = int(dtstr[5:7])

    if dtstr[7] != '-':
        raise ValueError('Invalid date separator')

    day = int(dtstr[8:10])

    return [year, month, day]


def _parse_hh_mm_ss_ff(tstr):
    # Parses things of the form HH[:MM[:SS[.fff[fff]]]]
    len_str = len(tstr)

    time_comps = [0, 0, 0, 0]
    pos = 0
    for comp in range(0, 3):
        if (len_str - pos) < 2:
            raise ValueError('Incomplete time component')

        time_comps[comp] = int(tstr[pos:pos+2])

        pos += 2
        next_char = tstr[pos:pos+1]

        if not next_char or comp >= 2:
            break

        if next_char != ':':
            raise ValueError('Invalid time separator: %c' % next_char)

        pos += 1

    if pos < len_str:
        if tstr[pos] != '.':
            raise ValueError('Invalid microsecond component')
        else:
            pos += 1

            len_remainder = len_str - pos
            if len_remainder not in (3, 6):
                raise ValueError('Invalid microsecond component')

            time_comps[3] = int(tstr[pos:])
            if len_remainder == 3:
                time_comps[3] *= 1000

    return time_comps


def _parse_isoformat_time(tstr):
    # Format supported is HH[:MM[:SS[.fff[fff]]]][+HH:MM[:SS[.ffffff]]]
    len_str = len(tstr)
    if len_str < 2:
        raise ValueError('Isoformat time too short')

    # This is equivalent to re.search('[+-]', tstr), but faster
    tz_pos = max(tstr.find('-'), tstr.find('+'), tstr.find('Z'))
    timestr = tstr[:tz_pos] if tz_pos > 0 else tstr

    time_comps = _parse_hh_mm_ss_ff(timestr)

    tzi = None
    if tz_pos > 0:
        tzstr = tstr[tz_pos:]

        # Valid time zone strings are:
        # Z                     len: 1
        # [+-]HH:MM             len: 6
        # [+-]HH:MM:SS          len: 9
        # [+-]HH:MM:SS.ffffff   len: 16

        if len(tzstr) not in (1, 6, 9, 16):
            raise ValueError('Malformed time zone string')

        if tzstr == 'Z':
            tzi = datetime.timezone.utc
        else:
            tz_comps = _parse_hh_mm_ss_ff(tzstr[1:])
            if all(x == 0 for x in tz_comps):
                tzi = datetime.timezone.utc
            else:
                tzsign = -1 if tstr[tz_pos] == '-' else 1

                td = datetime.timedelta(hours=tz_comps[0], minutes=tz_comps[1],
                                        seconds=tz_comps[2], microseconds=tz_comps[3])

                tzi = datetime.timezone(tzsign * td)

    time_comps.append(tzi)

    return time_comps


def fromisoformat(date_string):
    """Construct a datetime from the output of datetime.isoformat()."""
    if not isinstance(date_string, str):
        raise TypeError('fromisoformat: argument must be str')

    # Split this at the separator
    dstr = date_string[0:10]
    tstr = date_string[11:]

    try:
        date_components = _parse_isoformat_date(dstr)
    except ValueError:
        raise ValueError(f'Invalid isoformat string: {date_string!r}')

    if tstr:
        try:
            time_components = _parse_isoformat_time(tstr)
        except ValueError:
            raise ValueError(f'Invalid isoformat string: {date_string!r}')
    else:
        time_components = [0, 0, 0, 0, None]

    return datetime.datetime(*(date_components + time_components))



from unittest import TestCase

from wrgl.commit import Table
from wrgl.coldiff import longest_increasing_list, moveOps, MoveOp, ColDiff, Column, Move


class LongestIncreasingListTestCase(TestCase):
    def test_run(self):
        for l, res in [
            ([], []),
            ([0], [0]),
            ([0, 1], [0, 1]),
            ([1, 0], [0]),
            ([2, 0, 1], [1, 2]),
            ([1, 2, 0], [0, 1]),
            ([1, 0, 2], [0, 2]),
            ([0, 1, 2], [0, 1, 2]),
            ([2, 1, 0], [1]),
            ([0, 4, 5, 1, 2, 3], [0, 3, 4, 5]),
            ([0, 4, 5, 1, 2], [0, 1, 2]),
            ([4, 5, 0, 2, 1, 3], [2, 3, 5]),
        ]:
            self.assertEqual(longest_increasing_list(l), res)


class MoveOpsTestCase(TestCase):
    def test_run(self):
        for sl, ops in [
            ([], []),
            ([0], []),
            ([0, 1], []),
            ([0, 1, 2], []),
            ([1, 0], [MoveOp(old_ind=0, new_ind=1)]),
            ([1, 0, 2], [MoveOp(old_ind=0, new_ind=1)]),
            ([1, 2, 0], [MoveOp(old_ind=0, new_ind=2)]),
            ([2, 1, 0], [
                MoveOp(old_ind=2, new_ind=0),
                MoveOp(old_ind=0, new_ind=2),
            ]),
            ([0, 1, 2, 5, 3, 4], [MoveOp(old_ind=5, new_ind=3)]),
            ([2, 1, 4, 5, 3, 0], [
                MoveOp(old_ind=1, new_ind=1),
                MoveOp(old_ind=3, new_ind=4),
                MoveOp(old_ind=0, new_ind=5),
            ]),
        ]:
            self.assertEqual(moveOps(sl), ops)


class ColDiffTestCase(TestCase):
    def test_init(self):
        self.maxDiff = None
        for tbl_a, tbl_b, columns in [
            (
                Table(columns=["a"], pk=[]),
                Table(columns=["a"], pk=[]),
                [Column(name="a", base_idx=0, layer_idx={0: 0})]
            ),
            (
                Table(columns=["a"], pk=[]),
                Table(columns=["b"], pk=[]),
                [
                    Column(
                        name="a", base_idx=0, layer_idx={}, removed=set([0])
                    ),
                    Column(name="b", layer_idx={0: 0}, added=set([0]))
                ]
            ),
            (
                Table(columns=["a", "b"], pk=[]),
                Table(columns=["a", "b"], pk=[]),
                [
                    Column(name="a", base_idx=0, layer_idx={0: 0}),
                    Column(name="b", base_idx=1, layer_idx={0: 1})
                ]
            ),
            (
                Table(columns=["a", "b"], pk=[]),
                Table(columns=["b", "a"], pk=[]),
                [
                    Column(name="b", base_idx=1, layer_idx={0: 0}),
                    Column(
                        name="a", base_idx=0, layer_idx={0: 1}, moved={0: Move(before=0)}
                    )
                ]
            ),
            (
                Table(columns=["a", "b", "c"], pk=[]),
                Table(columns=["c", "a"], pk=[]),
                [
                    Column(name="c", base_idx=2, layer_idx={0: 0}),
                    Column(
                        name="a", base_idx=0, layer_idx={0: 1}, moved={0: Move(before=0)}
                    ),
                    Column(name="b", base_idx=1, removed=set([0]))
                ]
            ),
            (
                Table(columns=["c", "b", "a"], pk=[]),
                Table(columns=["a", "b", "c"], pk=[]),
                [
                    Column(
                        name="a", base_idx=2, layer_idx={0: 0}, moved={0: Move(after=1)}
                    ),
                    Column(name="b", base_idx=1, layer_idx={0: 1}),
                    Column(
                        name="c", base_idx=0, layer_idx={0: 2}, moved={0: Move(before=1)}
                    )
                ]
            ),
            (
                Table(columns=["a", "b"], pk=[]),
                Table(columns=["b", "c", "a"], pk=[]),
                [
                    Column(name="b", base_idx=1, layer_idx={0: 0}),
                    Column(name="c", layer_idx={0: 1}, added=set([0])),
                    Column(
                        name="a", base_idx=0, layer_idx={0: 2}, moved={0: Move(before=0)}
                    )
                ]
            ),
            (
                Table(columns=["a", "d", "e", "b", "c"], pk=[]),
                Table(columns=["a", "b", "c", "d", "e"], pk=[]),
                [
                    Column(name="a", base_idx=0, layer_idx={0: 0}),
                    Column(name="b", base_idx=3, layer_idx={0: 1}),
                    Column(name="c", base_idx=4, layer_idx={0: 2}),
                    Column(
                        name="d", base_idx=1, layer_idx={0: 3}, moved={0: Move(after=0)}
                    ),
                    Column(
                        name="e", base_idx=2, layer_idx={0: 4}, moved={0: Move(after=0)}
                    )
                ]
            ),
            (
                Table(columns=["e", "b", "c", "d", "f"], pk=[]),
                Table(columns=["a", "b", "c", "d", "e"], pk=[]),
                [
                    Column(name="a", layer_idx={0: 0}, added=set([0])),
                    Column(name="b", base_idx=1, layer_idx={0: 1}),
                    Column(name="c", base_idx=2, layer_idx={0: 2}),
                    Column(name="d", base_idx=3, layer_idx={0: 3}),
                    Column(name="f", base_idx=4, removed=set([0])),
                    Column(
                        name="e", base_idx=0, layer_idx={0: 4}, moved={0: Move(before=1)}
                    )
                ]
            ),
            (
                Table(columns=["a", "b", "c"], pk=[0]),
                Table(columns=["a", "b", "c"], pk=[0]),
                [
                    Column(
                        base_idx=0, base_pk=True, layer_idx={0: 0}, layer_pk=set([0]), name="a"
                    ),
                    Column(
                        base_idx=1, layer_idx={0: 1}, name="b"
                    ),
                    Column(
                        base_idx=2, layer_idx={0: 2}, name="c"
                    )
                ]
            ),
            (
                Table(columns=["a", "b", "c"], pk=[0]),
                Table(columns=["b", "a", "d"], pk=[2, 1]),
                [
                    Column(
                        added=set([0]), layer_idx={0: 2}, layer_pk=set([0]), name="d"
                    ),
                    Column(
                        base_idx=0, base_pk=True, layer_idx={0: 1}, layer_pk=set([0]), moved={0: Move(before=0)}, name="a"
                    ),
                    Column(
                        base_idx=1, layer_idx={0: 0}, name="b"
                    ),
                    Column(
                        base_idx=2, name="c", removed=set([0])
                    )
                ]
            ),
            (
                Table(
                    columns=[
                        "a", "ab", "ac", "ad", "b", "c", "ca", "cb", "cd", "e", "ea", "eb", "ec",
                    ],
                    pk=[]
                ),
                Table(columns=["a", "b", "c", "d", "e", "f"], pk=[]),
                [
                    Column(name="a", base_idx=0, layer_idx={0: 0}),
                    Column(name="ab", base_idx=1, removed=set([0])),
                    Column(name="ac", base_idx=2, removed=set([0])),
                    Column(name="ad", base_idx=3, removed=set([0])),
                    Column(name="b", base_idx=4, layer_idx={0: 1}),
                    Column(name="c", base_idx=5, layer_idx={0: 2}),
                    Column(name="ca", base_idx=6, removed=set([0])),
                    Column(name="cb", base_idx=7, removed=set([0])),
                    Column(name="cd", base_idx=8, removed=set([0])),
                    Column(name="d", added=set([0]), layer_idx={0: 3}),
                    Column(name="e", base_idx=9, layer_idx={0: 4}),
                    Column(name="ea", base_idx=10, removed=set([0])),
                    Column(name="eb", base_idx=11, removed=set([0])),
                    Column(name="ec", base_idx=12, removed=set([0])),
                    Column(name="f", added=set([0]), layer_idx={0: 5})
                ]
            )
        ]:
            cd = ColDiff(tbl_a, tbl_b)
            self.assertEqual(cd.columns, columns)

    def test_rearrange_base_row(self):
        for tbl_a, tbl_b, row, rearranged in [
            (
                Table(columns=["a", "b", "c"], pk=[0]),
                Table(columns=["a", "b", "c"], pk=[0]),
                ["1", "2", "3"],
                ["1", "2", "3"]
            ),
            (
                Table(columns=["a", "b", "c"], pk=[0]),
                Table(columns=["b", "a", "d"], pk=[2, 1]),
                ["1", "2", "3"],
                [None, "1", "2", "3"]
            )
        ]:
            cd = ColDiff(tbl_a, tbl_b)
            self.assertEqual(cd.rearrange_base_row(row), rearranged)

    def test_rearrange_row(self):
        for tbl_a, tbl_b, row, rearranged in [
            [
                Table(columns=["a", "b", "c"], pk=[0]),
                Table(columns=["a", "b", "c"], pk=[0]),
                ["1", "2", "3"],
                ["1", "2", "3"]
            ],
            [
                Table(columns=["a", "b", "c"], pk=[0]),
                Table(columns=["b", "a", "d"], pk=[2, 1]),
                ["1", "2", "3"],
                ["3", "2", "1", None]
            ]
        ]:
            cd = ColDiff(tbl_a, tbl_b)
            self.assertEqual(cd.rearrange_row(0, row), rearranged)

    def test_combine_rows(self):
        cd = ColDiff(
            Table(columns=["e", "b", "c", "d", "f"]),
            Table(columns=["a", "b", "c", "d", "e"])
        )
        self.assertEqual(
            cd.columns,
            [
                Column(name="a", added=set([0]), layer_idx={0: 0}),
                Column(name="b", layer_idx={0: 1}, base_idx=1),
                Column(name="c", layer_idx={0: 2}, base_idx=2),
                Column(name="d", layer_idx={0: 3}, base_idx=3),
                Column(name="f", removed=set([0]), base_idx=4),
                Column(
                    name="e", layer_idx={0: 4}, base_idx=0, moved={0: Move(before=1)}
                )
            ]
        )
        self.assertEqual(
            cd.combine_rows(
                0,
                ["1", "2", "3", "4", "5"],
                ["6", "2", "7", "4", "5"]
            ),
            [
                ("1", None), ("2", "2"), ("3", "7"),
                ("4", "4"), (None, "5"), ("5", "6")
            ]
        )

    def test_no_column_changes(self):
        for tbl_a, tbl_b, result in [
            (
                Table(columns=["a", "b", "c"]),
                Table(columns=["a", "b", "c"]),
                True
            ),
            (
                Table(columns=["a", "b", "c"]),
                Table(columns=["a", "b", "c", "d"]),
                False
            ),
            (
                Table(columns=["a", "b", "c", "d"]),
                Table(columns=["a", "b", "c"]),
                False
            ),
            (
                Table(columns=["a", "b", "c"]),
                Table(columns=["a", "c", "b"]),
                False
            )
        ]:
            cd = ColDiff(tbl_a, tbl_b)
            self.assertEqual(cd.no_column_changes(), result)

    def test_col_status(self):
        cd = ColDiff(
            Table(columns=["a", "b", "c", "e"]),
            Table(columns=["a", "e", "b", "d"]),
        )
        self.assertEqual(cd.columns[0].name, "a")
        self.assertEqual(cd.columns[0].is_added, False)
        self.assertEqual(cd.columns[0].is_removed, False)
        self.assertEqual(cd.columns[0].is_moved, False)

        self.assertEqual(cd.columns[2].name, "b")
        self.assertEqual(cd.columns[2].is_added, False)
        self.assertEqual(cd.columns[2].is_removed, False)
        self.assertEqual(cd.columns[2].is_moved, True)

        self.assertEqual(cd.columns[3].name, "c")
        self.assertEqual(cd.columns[3].is_added, False)
        self.assertEqual(cd.columns[3].is_removed, True)
        self.assertEqual(cd.columns[3].is_moved, False)

        self.assertEqual(cd.columns[4].name, "d")
        self.assertEqual(cd.columns[4].is_added, True)
        self.assertEqual(cd.columns[4].is_removed, False)
        self.assertEqual(cd.columns[4].is_moved, False)



import re
from typing import Union, Dict, Tuple, Callable

import requests
from requests.exceptions import HTTPError

from wrgl.serialize import json_dumps


class UMAClient:
    _rsc_uri: str = ""
    _client_id: str = ""
    _client_secret: str = ""
    _access_token: str = ""
    rpt: str = ""

    def __init__(self, rsc_uri: str, client_id: str, client_secret: str) -> None:
        """
        :param str rsc_uri: the URI of the UMA resource
        :param str client_id: Keycloak client id
        :param str client_secret: Keycloak client secret
        """
        self._rsc_uri = rsc_uri.rstrip("/")
        self._client_id = client_id
        self._client_secret = client_secret

    def _headers(self, headers=None) -> Union[Dict, None]:
        if self.rpt:
            if headers is None:
                headers = dict()
            headers["Authorization"] = "Bearer " + self.rpt
        return headers

    def _extract_uma_ticket(
        self, resp: requests.Response
    ) -> Tuple[Union[str, None], Union[str, None]]:
        auth_header = resp.headers.get("WWW-Authenticate", "")
        if auth_header.startswith("UMA "):
            parts = auth_header.split(",")
            values = dict()
            for s in parts:
                key, value = tuple(s.strip().split("="))
                values[key] = value.strip('"')
            return values.get("as_uri", None), values.get("ticket", None)
        return None, None

    def _discover_uma_config(self, as_uri: str) -> Dict:
        resp = requests.get(as_uri.rstrip("/") + "/.well-known/uma2-configuration")
        resp.raise_for_status()
        return resp.json()

    def _fetch_access_token(self, token_endpoint: str) -> None:
        resp = requests.post(
            token_endpoint,
            data={
                "grant_type": "client_credentials",
                "client_id": self._client_id,
                "client_secret": self._client_secret,
            },
        )
        resp.raise_for_status()
        resp_data = resp.json()
        self._access_token = resp_data["access_token"]
        return

    def _fetch_rpt(self, token_endpoint: str, uma_ticket: str) -> None:
        resp = requests.post(
            token_endpoint,
            data={
                "grant_type": "urn:ietf:params:oauth:grant-type:uma-ticket",
                "ticket": uma_ticket,
            },
            headers={"Authorization": "Bearer " + self._access_token},
        )
        resp.raise_for_status()
        resp_data = resp.json()
        self.rpt = resp_data["access_token"]
        return

    def _ensure_rpt(self, as_uri: str, uma_ticket: str) -> None:
        uma_config = self._discover_uma_config(as_uri)
        token_endpoint = uma_config["token_endpoint"]
        if not self._access_token:
            self._fetch_access_token(token_endpoint)
        try:
            self._fetch_rpt(token_endpoint, uma_ticket)
        except HTTPError as e:
            if e.response.status_code == 401:
                self._fetch_access_token(token_endpoint)
                self._fetch_rpt(token_endpoint, uma_ticket)
            else:
                raise

    def _do_request(
        self,
        method: str,
        url: str,
        params=None,
        headers=None,
        create_request_args: Union[Callable[..., Dict], None] = None,
        *args,
        **kwargs
    ) -> requests.Response:
        if create_request_args is not None:
            args_dict = create_request_args()
            args_dict["headers"] = self._headers(args_dict.get("headers", None))
            return requests.request(method, url, **args_dict)
        if params is not None:
            params = {k: v for k, v in params.items() if v}
        return requests.request(
            method, url, params=params, headers=self._headers(headers), *args, **kwargs
        )

    def request(
        self,
        method: str,
        path: str,
        params=None,
        headers=None,
        create_request_args: Union[Callable[..., Dict], None] = None,
        *args,
        **kwargs
    ) -> requests.Response:
        """Make a request

        :param str method: HTTP verb
        :param str path: path relative to rsc_uri
        :param dict params: optional, query parameters
        :param dict headers: optional, HTTP headers
        :param func create_request_args: optional. If defined, this function is called to generate request arguments. Useful when request need to be retried.
        :param list args: extra positional arguments passed to requests.request. Ignored if create_request_args is defined.
        :param dict kwargs: extra keyword arguments passed to requests.request. Ignored if create_request_args is defined.

        :rtype: requests.Response
        """
        url = self._rsc_uri + path
        resp = self._do_request(
            method, url, params, headers, create_request_args, *args, **kwargs
        )
        if resp.status_code == 401:
            as_uri, uma_ticket = self._extract_uma_ticket(resp)
            if uma_ticket is not None:
                self._ensure_rpt(as_uri, uma_ticket)
                resp = self._do_request(
                    method, url, params, headers, create_request_args, *args, **kwargs
                )
        resp.raise_for_status()
        return resp

    def get(
        self, path: str, params=None, headers=None, *args, **kwargs
    ) -> requests.Response:
        """Make a get request

        :param str path: path relative to rsc_uri
        :param dict params: optional, query parameters
        :param dict headers: optional, HTTP headers
        :param list args: extra positional arguments passed to requests.request. Ignored if create_request_args is defined.
        :param dict kwargs: extra keyword arguments passed to requests.request. Ignored if create_request_args is defined.

        :rtype: requests.Response
        """
        return self.request(
            "GET", path, params=params, headers=headers, *args, **kwargs
        )

    def post(
        self,
        path: str,
        params=None,
        headers=None,
        create_request_args: Union[Callable[..., Dict], None] = None,
        *args,
        **kwargs
    ) -> requests.Response:
        """Make a post request

        :param str path: path relative to rsc_uri
        :param dict params: optional, query parameters
        :param dict headers: optional, HTTP headers
        :param func create_request_args: optional. If defined, this function is called to generate request arguments. Useful when request need to be retried.
        :param list args: extra positional arguments passed to requests.request. Ignored if create_request_args is defined.
        :param dict kwargs: extra keyword arguments passed to requests.request. Ignored if create_request_args is defined.

        :rtype: requests.Response
        """
        return self.request(
            "POST",
            path,
            params=params,
            headers=headers,
            create_request_args=create_request_args,
            *args,
            **kwargs
        )



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2022 Wrangle Ltd

import typing
from unittest import TestCase

import attr

from wrgl.serialize import json_loads, json_dumps, field_transformer


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Person(object):
    name: str
    birth_date: str
    height: int
    scores: typing.List[int]


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Remote(object):
    url: str


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Branch(object):
    fetch: str


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Config(object):
    user: Person
    remotes: typing.List[Remote]
    branch: typing.Dict[str, Branch]


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Receive(object):
    deny_non_fast_forwards: bool
    deny_deletes: bool


class JSONTestCase(TestCase):
    def test_loads_simple(self):
        obj1 = Person(
            name="John Doe", birth_date="10/02/2000", height=170, scores=[7, 8, 9]
        )
        json_str = json_dumps(obj1)
        self.assertEqual(
            json_str,
            '{"name": "John Doe", "birthDate": "10/02/2000", "height": 170, "scores": [7, 8, 9]}',
        )
        obj2 = json_loads(json_str, Person)
        self.assertEqual(obj1, obj2)

    def test_loads_with_null(self):
        self.assertEqual(
            json_loads(
                '{"user": null, "remotes": null, "branch": null, "nonExistentKey": null}',
                Config,
            ),
            Config(),
        )

    def test_loads_nested(self):
        self.maxDiff = None
        obj1 = Config(
            user=Person(name="John Doe", birth_date="10/02/2000", height=170),
            remotes=[
                Remote(url="https://url1"),
                Remote(url="https://url2"),
            ],
            branch={
                "main": Branch(fetch="refs/heads/main"),
                "test": Branch(fetch="refs/heads/test"),
            },
        )
        json_str = json_dumps(obj1, indent=4)
        self.assertEqual(
            json_str,
            "\n".join(
                [
                    "{",
                    '    "user": {',
                    '        "name": "John Doe",',
                    '        "birthDate": "10/02/2000",',
                    '        "height": 170',
                    "    },",
                    '    "remotes": [',
                    "        {",
                    '            "url": "https://url1"',
                    "        },",
                    "        {",
                    '            "url": "https://url2"',
                    "        }",
                    "    ],",
                    '    "branch": {',
                    '        "main": {',
                    '            "fetch": "refs/heads/main"',
                    "        },",
                    '        "test": {',
                    '            "fetch": "refs/heads/test"',
                    "        }",
                    "    }",
                    "}",
                ]
            ),
        )
        obj2 = json_loads(json_str, Config)
        self.assertEqual(obj1, obj2)

    def test_loads_ignore_empty(self):
        obj = json_loads('{"denyNonFastForwards": true}', Receive)
        self.assertEqual(obj, Receive(deny_non_fast_forwards=True))
